import { LogLevel } from '../types';
export declare class Logger {
    private readonly logLevel;
    constructor(level?: LogLevel);
    /**
     * Calls console.error with provided information, will always be called
     *
     * @param {string} className Class where error originated from
     * @param {string} error Error message that will displayed in the console
     * @param {object} additionalData Optional additional data that will be shown in console
     */
    error(className: string, error: string, additionalData?: any): void;
    /**
     * Calls console.warn with provided information,
     * will only be called if LogLevel is Debug, Info or Warning
     *
     * @param {string} className Class where warning originated from
     * @param {string} warning Warning message that will displayed in the console
     * @param {object} additionalData Optional additional data that will be shown in console
     */
    warn(className: string, warning: string, additionalData?: any): void;
    /**
     * Calls console.info with provided information,
     * will only be called if LogLevel is Debug or Info
     *
     * @param {string} className Class where info message originated from
     * @param {string} message Info message that will displayed in the console
     * @param {object} additionalData Optional additional data that will be shown in console
     */
    info(className: string, message: string, additionalData?: any): void;
    /**
     * Calls console.debug with provided information,
     * will only be called if LogLevel is Debug
     *
     * @param {string} className Class where debug message originated from
     * @param {string} message Debug message that will displayed in the console
     * @param {object} additionalData Optional additional data that will be shown in console
     */
    debug(className: string, message: string, additionalData?: any): void;
}
