import { ClientOptions } from '../types';
import { Logger } from './logger';
interface PkceArtifacts {
    codeVerifier: string;
    codeChallenge?: string;
}
export declare class OAuth {
    /**
     *
     * @param str
     * @returns
     */
    static btoa(encodeStr: string): string;
    static atob(decodeStr: string): string;
    /**
     * Generates the code_challenge parameter to support the PKCE workflow.
     *
     * @param {string} codeVerifier Used to validate already received code_challenge
     * @returns {string} code challenge that has been generated
     */
    static generateCodeChallenge(codeVerifier: string): Promise<string>;
    /**
      Generates a random string used for state and PKCE code_challenge.
  
      @param {number} length Length of the generated string
      @return {string} random string that has been generated
      */
    static getRandomString(length: number): string;
    /**
     * Generates the artifacts necessary to add code challenge parameters to request
     *
     * @param {ClientOptions} options Options passed into the authorize method
     * @param {Logger} logger instance of the logger set up calling class
     * @returns {Promise<PkceArtifacts>} artifacts necessary to be sent through the request body or URL
     */
    static generatePkceArtifacts(options: ClientOptions, logger: Logger): Promise<PkceArtifacts>;
    static getOrGenerateState(options: ClientOptions, logger: Logger): string;
}
export default OAuth;
