/**
 * Wrapper subclass representing a Web Worker for in-memory storage in a separate thread.
 * This has to be a wrapper class because a Web Worker cannot access the main thread and vice versa.
 * @see worker-thread.js
 * {@link https://developer.mozilla.org/en-US/docs/Web/API/Web_Workers_API/Functions_and_classes_available_to_workers}
 */
import { TokenResponse } from '../types/token-response';
import { ClientStorageBase } from './client-storage-base';
export interface WorkerMessage {
    method: string;
    payload?: string | object;
}
export declare class WorkerClientStorage extends ClientStorageBase {
    private readonly workerThread;
    private msg;
    constructor(clientId: string);
    storeToken(token: TokenResponse): void;
    getToken(): Promise<TokenResponse>;
    getRefreshToken(): Promise<string>;
    removeToken(): void;
}
